/*
 * cc -I/usr/include/drm -ldrm drift_control.c -o drift_control
 *
 * drift_control:
 *   - without args simply monitors drift of frame rate against loop cycle
 *   - otherwise adjusts frame rate to loop cycle frequency +-0.001441383Hz
 *
 * prerequisites:
 *   - Xserver must be running 
 */
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <xf86drm.h>
#include <radeon_drm.h>

#define min(a, b) ((a) <= (b) ? (a) : (b))
#define max(a, b) ((a) >= (b) ? (a) : (b))

#define STANDALONE

/* --- 8< --- */
/*
 * we average 25 frames to yield a cycle time of
 * about one second for analysis of frame rate data.
 * this serves as frequency divider for our software PLL.
 */
#ifdef STANDALONE
#define SYF_PLL_DIVIDER 10
#else
#define SYF_PLL_DIVIDER 25
#endif

/*
 * offset in usecs from double buffer switch where we try to place double
 * buffer updates.
 * this lowers sensivity to jitter of our software PLL phase comparator.
 */
#define SYF_SYNC_POINT 20000

/*
 * one trim increment compensates drift speed for about 29usec/sec.
 * this represents resolution of our VCO input. 
 */
#define SYF_MIN_STEP_USEC 100

/*
 * factor weighting drift against sync point displacement
 * when calculating overall compensation 
 */
#define SYF_DISP_DRIFT_FACTOR 1

/* NOT CURRENTLY USED
 * to allow for smooth adaption also on slower devices we delimit maximum
 * trim change per step size.
 * this implements some kind of low pass filter for our VCO input.
 */
#define SYF_MAX_TRIM_REL 1000

/*
 * maximum absolute trim values allowed due to current
 * hardware/driver contraints.
 * this delimits 'maximum voltage' controlling our VCO.
 */
#define SYF_MAX_TRIM_ABS 37

#define DEBUG
#ifdef DEBUG
#define ERRORF ErrorF
#else
#define ERRORF(...)
#endif

#ifdef STANDALONE
#define ErrorF printf
#define B(a) (*(argv + (a)) ? *(argv + (a)) : 0)
#else
#define B(a) (a)
#endif

//#define USE_METER
#ifdef USE_METER
#define OUT_ERRORF(...)
#define OUT_GRAPHIC meter_out
#else
#define OUT_ERRORF ErrorF
#define OUT_GRAPHIC(...)
#endif

void
meter_out(val, symb)
{
    static char meter[81];

    if (!symb) {
        ErrorF("%s\n", meter);
        memset(meter, '-', 80);
        return;
    }
    val /= 500;
    val = min(val,  39);
    val = max(val, -40);
    meter[40 + val] = symb;
}

/* --- 8< --- */

main(argc, argv)
    char **argv;
{
    static int fd;
    static int cnt;
    static int trim;
    static int ds_usecs;
    static int sync_point_disp;
    static struct timeval skew_prev = { ~0 };
    drm_radeon_syncf_t syncf;
    drm_radeon_setparam_t vbl_activate;
    struct timeval skew;
    struct timeval drift_speed;
    int tmp;

    if ((fd = drmOpen("radeon", 0)) < 0) {
	ErrorF("drmOpen: %s\n", strerror(errno));
    }
    vbl_activate.param = RADEON_SETPARAM_VBLANK_CRTC;
    vbl_activate.value = DRM_RADEON_VBLANK_CRTC1;
    if (ioctl(fd, DRM_IOCTL_RADEON_SETPARAM, &vbl_activate)) {
        ErrorF("DRM_IOCTL_RADEON_SETPARAM: %s\n", strerror(errno));
    }
    while (1) {
        syncf.trim = 0;
        if (ioctl(fd, DRM_IOCTL_RADEON_SYNCF, &syncf)) {
            ErrorF("DRM_IOCTL_RADEON_SYNCF: %s\n", strerror(errno));
        }

/* --- 8< --- */
#ifndef STANDALONE
	if (syncf.vbls & 1) {
	    struct drm_wait_vblank_request vbr;
loop:
	    vbr.type = _DRM_VBLANK_ABSOLUTE;
	    vbr.sequence = syncf.vbls + 1;
	    vbr.signal = 0;
	    if (ioctl(fd, DRM_IOCTL_WAIT_VBLANK, &vbr)) {
		ErrorF("DRM_IOCTL_WAIT_VBLANK: %s\n", strerror(errno));
	    }
            if (((struct drm_wait_vblank_reply *)&vbr)->sequence < syncf.vbls + 1) {
		/* AFTER EINTR */
                goto loop;
            }
	}
#endif
	VSF_SUB(syncf.tv_now, syncf.tv_vbl, skew)
	if (skew_prev.tv_sec != ~0) {
	    sync_point_disp += skew.tv_usec - SYF_SYNC_POINT;
	    VSF_SUB(skew, skew_prev, drift_speed);
	    VSF_TV2USEC(drift_speed, tmp);
	    ds_usecs += tmp;
	    ++cnt;
	}
	if (skew_prev.tv_sec != ~0 && !(cnt % SYF_PLL_DIVIDER)) {
	    sync_point_disp /= SYF_PLL_DIVIDER;
	    ERRORF("\n");
	    ERRORF("tv now:           %10d.%06d\n", (int)syncf.tv_now.tv_sec, (int)syncf.tv_now.tv_usec);
	    ERRORF("tv vbl:           %10d.%06d\n", (int)syncf.tv_vbl.tv_sec, (int)syncf.tv_vbl.tv_usec);
	    ERRORF("vbls:                    %10d\n", syncf.vbls);
	    ERRORF("trim:                    0x%08x\n", syncf.trim);
            OUT_ERRORF("sync point displacement: %10d\n", sync_point_disp);
            OUT_ERRORF("drift speed:             %10d %s\n", ds_usecs, abs(ds_usecs) > 15000 ? "excessive drift speed" : "");
            OUT_GRAPHIC(0, '+');
            OUT_GRAPHIC(sync_point_disp, '|');
            OUT_GRAPHIC(ds_usecs, '*');
            OUT_GRAPHIC(0, 0);
	    if (B(1)) {
		trim = (ds_usecs + sync_point_disp / SYF_DISP_DRIFT_FACTOR) / SYF_MIN_STEP_USEC;
                OUT_ERRORF("overall compensation:    %10d %s\n", trim, abs(trim) <= 1 ? "completed" : "");
		trim = max(trim, -SYF_MAX_TRIM_REL);
		trim = min(trim,  SYF_MAX_TRIM_REL);
		ERRORF("o. c. clipped:           %10d\n", trim);
	    }
	    sync_point_disp = 0;
	    ds_usecs = 0;
	}
        if (trim) {
            tmp = (char)(syncf.trim & 0xff);
            if (trim > 0) {
                tmp = min(tmp + 1,  SYF_MAX_TRIM_ABS);
                --trim;
            } else {
                tmp = max(tmp - 1, -SYF_MAX_TRIM_ABS);
                ++trim;
            }
            syncf.trim = VSF_SET_TRIM | VSF_TEMPLATE | tmp & 0xff;
	    ERRORF("*trim:                   0x%08x\n", syncf.trim);
            if (ioctl(fd, DRM_IOCTL_RADEON_SYNCF, &syncf)) {
                ErrorF("DRM_IOCTL_RADEON_SYNCF: %s\n", strerror(errno));
            }
	}
        skew_prev = skew;
/* --- 8< --- */

        usleep( 99988);  /* OPT */
    }
}

